/*
 * Copyright (c) 2008-2010 Institut Geographique National France, released under the
 * BSD license.
 */
/*
 * @requires Geoportal/Format.js
 */
/**
 * Class: Geoportal.Format.GPX
 * Read/write GPX parser. Create a new instance with the
 *     <Geoportal.Format.GPX> constructor.
 *     @see <http://www.topografix.com/gpx.asp>
 *
 * Inherits from:
 *  - <Geoportal.Format>
 *  - <OpenLayers.Format.XML>
 */
Geoportal.Format.GPX = OpenLayers.Class(Geoportal.Format, OpenLayers.Format.XML, {

   /**
     * APIProperty: defaultVersion
     * {String} Version number to assume if none found.  Default is "1.0".
     */
    defaultVersion: "1.0",

    /**
     * APIProperty: version
     * {String} Specify a version string if one is known.
     */
    version: null,

    /**
     * Property: parser
     * {Object} Instance of the versioned parser.  Cached for multiple read and
     *     write calls of the same version.
     */
    parser: null,

    /**
     * Constructor: Geoportal.Format.GPX
     * Create a new parser for GPX.
     *
     * Parameters:
     * options - {Object} An optional object whose properties will be set on
     *     this instance.
     *      A mapping between GPX tag and feature's attributes can be passed
     *      in the options object. The default is to keep GPX tag names as
     *      attribute names. If a mapping is null, the GPX tag is skipped.
     *
     * (code)
     * var format= new Geoportal.Format.GPX({
     *      mapping:{
     *          'email': null,      // don't map email by default
     *          'geoidheight': 'H'  // map with 'H' by default
     *                              // any other GPX tag are read/written keeping
     *                              // the GPX name
     *          'gpx':{
     *              ...
     *          },
     *          'wpt':{
     *              'ele': 'z',     // 'ele' is mapped with 'z' attribute
     *              'url': null     // 'url' is never read/written
     *          }
     *      }
     * });
     * (end)
     */
    initialize: function(options) {
        options= options || {};
        options.mapping= OpenLayers.Util.extend({}, options.mapping);
        OpenLayers.Format.XML.prototype.initialize.apply(this, [options]);
    },

    /**
     * APIMethod: read
     * Return a list of features from a GPX doc
     * Parameters:
     * data - {String} or {DOMElement} data to read/parse.
     *
     * Returns:
     * {Array({<OpenLayers.Feature.Vector>})} that may be empty.
     */
    read: function(data) {
        if (typeof(data)=="string") {
            data= OpenLayers.Format.XML.prototype.read.apply(this, [data]);
        }
        if (!data) { return []; }
        var root= (data.nodeType==9 ? data.documentElement : data);
        var version= root.getAttribute('version') || this.version || this.defaultVersion;
        if (!this.parser || this.parser.VERSION != version) {
            var format= Geoportal.Format.GPX['v'+version.replace(/\./g, "_")];
            if (!format) {
                throw OpenLayers.i18n('GPX.version', { 'gpxVersion': version });
            }
            this.parser= new format(this.options);
        }
        var features= this.parser.read(data);
        this.version= this.parser.version;
        return features;
    },

    /**
     * APIMethod: write
     * Generate a GPX document string given a list of features.
     *
     * Parameters:
     * features - {Array(<OpenLayers.Feature.Vector>)} List of features to
     *     serialize into a string.
     * options - {Object} Optional configuration object.
     *
     * Returns:
     * {String} A string representing the GPX document.
     */
    write: function(features, options) {
        var version= (options && options.version) || this.version || this.defaultVersion;
        if (!this.parser || this.parser.VERSION != version) {
            var format= Geoportal.Format.GPX["v" + version.replace(/\./g, "_")];
            if (!format) {
                throw OpenLayers.i18n('GPX.version', { 'gpxVersion': version });
            }
            this.parser= new format(this.options);
            this.parser.tags= eval("Geoportal.Format.GPX.v"+version.replace(/\./g, "_")+".TAGS");
        }
        var xml= this.parser.write(features,options);
        this.version= this.parser.version;
        return xml;
    },

    /**
     * Constant: CLASS_NAME
     * {String} *"Geoportal.Format.GPX"*
     */
    CLASS_NAME: "Geoportal.Format.GPX"
});
